#! /usr/bin/perl
use strict;
use warnings;
no warnings qw(once);
use CGI;
use CGI::Carp qw( fatalsToBrowser );
use DBI;
use POSIX;
use Devel::Peek;
use HTML::Entities;
use Encode;
use locale;
require '../config.pm';

my $cgi = CGI->new;

my $dbh = DBI->connect($config::local_connstr, $config::local_username, $config::local_password)
	or die "connect: " . $DBI::errstr;
$dbh->{AutoCommit} = 0;
$dbh->{RaiseError} = 1;

my $trials = 25_000;

binmode STDOUT, ':utf8';

my %players = ();
my %ratings = ();
my @matches = ();
my %parms = ();
my $match_stddev;

sub sanitize {
	return HTML::Entities::encode_entities(shift);
}

sub color {
	my $x = shift;
	return int(255.0 * ($x ** (1.80)));
}

sub make_table {
	my ($lowest_division, $used_ratings) = @_;

	print <<"EOF";

    <table>
      <tr>
        <th></th>
EOF

	POSIX::setlocale(&POSIX::LC_ALL, 'C');

	my $tmpnam = POSIX::tmpnam();
	open MCCALC, ">", $tmpnam
		or die "$tmpnam: $!";

	printf MCCALC "%f\n", $match_stddev;
	printf MCCALC "%d\n", scalar keys %players;

	for my $id (keys %players) {
		if (exists($used_ratings->{$id}) && defined($used_ratings->{$id})) {
			printf MCCALC "%s %f\n", $id, $used_ratings->{$id};
		} else {
			printf MCCALC "%s %f\n", $id, 1500.0;
		}
	}

	for my $match (@matches) {
		printf MCCALC "%s %s %d %d\n", $match->[0], $match->[1], $match->[2], $match->[3];
	}
	close MCCALC;

	POSIX::setlocale(&POSIX::LC_ALL, 'nb_NO.UTF-8');

	my %prob = ();

	open MCCALC, "$config::base_dir/mcwordfeud $trials < $tmpnam |"
		or die "mccalc: $!";
	while (<MCCALC>) {
		chomp;
		my @x = split /\s+/;
		my $id = $x[0];
		my $player = sprintf "%s (%.0f)", $players{$id}, $ratings{$id};
		$prob{$player} = [ @x[1..$#x] ];
	}
	close MCCALC;
	#unlink $tmpnam;

	my $num_games = scalar keys %prob;
	for my $i (1..$num_games) {
		print "        <th>$i.</th>\n";
	}
	print "        <th>NEDRYKK</th>\n" unless ($lowest_division);
	print "      </tr>\n";

	for my $player (sort { $a cmp $b } keys %prob) {
		print "      <tr>\n";
		print "        <th>$player</th>\n";

		for my $i (1..$num_games) {
			my $pn = $prob{$player}->[$i - 1] / $trials;

			my $r = color(1.0 - $pn / 3);
			my $g = color(1.0 - $pn / 3);
			my $b = color(1.0);

			if ($i == 1) {
				($g, $b) = ($b, $g);
			} elsif ($i >= $num_games - 1 && !$lowest_division) {
				($r, $b) = ($b, $r);
			}

			printf "        <td style=\"background-color: rgb($r, $g, $b)\" class=\"num\">%.1f%%</td>\n", $pn * 100.0;
		}

		unless ($lowest_division) {
			my $pn = ($prob{$player}->[$num_games - 1] + $prob{$player}->[$num_games - 2]) / $trials;

			my $r = color(1.0);
			my $g = color(1.0 - $pn / 3);
			my $b = color(1.0 - $pn / 3);
			printf "        <td style=\"background-color: rgb($r, $g, $b)\" class=\"num\">%.1f%%</td>\n", $pn * 100.0;
		}
		print "      </tr>\n";
	}

	print << "EOF";
    </table>
EOF
}

# Get auxillary parameters
my $q = $dbh->prepare('SELECT * FROM ratings WHERE id < 0');
$q->execute;

while (my $ref = $q->fetchrow_hashref) {
	$parms{$ref->{'id'}} = $ref->{'rating'};
}
$match_stddev = $parms{-2} * sqrt(2.0);

my $season;
my $division = $cgi->param('divisjon') // -1;
my $subdivision = $cgi->param('avdeling') // -1;
my $last_division = 0;

POSIX::setlocale(&POSIX::LC_ALL, 'nb_NO.UTF-8');

print $cgi->header(-type=>'text/html; charset=utf-8', -expires=>'now');
printf <<"EOF", $match_stddev;
<html>
  <head>
    <title>WLoH-plasseringsannsynlighetsberegning</title>
    <link rel="stylesheet" href="/style" type="text/css" />
  </head>
  <body>
    <h1>WLoH-plasseringsannsynlighetsberegning</h1>

    <p><em>Dette er et hobbyprosjekt fra tredjepart, og ikke en offisiell del av
      <a href="http://wordfeud.aasmul.net/">Wordfeud Leage of Honour</a>.</em></p>

    <p>Beregningen tar ikke hensyn til ujevn spillestyrke, ting som er sagt i forumet e.l.;
      den antar at samtlige uspilte kamper trekkes fra en normalfordeling med standardavvik
      %.1f poeng. Sannsynlighetene kan summere til andre tall enn 100%% pga. avrunding.
      Tallene vil variere litt fra gang til gang fordi utregningen skjer ved randomisering.</p>

    <p>Spillerne er sortert etter nick.</p>

    <form method="get" action="/">
EOF

$q = $dbh->prepare('SELECT MAX(sesong) AS max_sesong FROM fotballserier');
$q->execute;
my $season;
while (my $ref = $q->fetchrow_hashref) {
	$season = $ref->{'max_sesong'};
}

print <<"EOF";
     <p>Divisjon:
        <select name="divisjon" onchange="form.submit();">
EOF

$q = $dbh->prepare('SELECT DISTINCT(divisjon) FROM fotballserier WHERE sesong=? ORDER BY divisjon');
$q->execute($season);

my $found_division = 0;
my $max_division;

while (my $ref = $q->fetchrow_hashref) {
	my $d = $ref->{'divisjon'};
	if ($d == $division) {
		print "        <option value=\"$d\" selected=\"selected\">$d</option>\n";
		$found_division = 1;
	} else {
		print "        <option value=\"$d\">$d</option>\n";
	}
	$max_division = $d;
}

$division = 1 if (!$found_division);

print <<"EOF";
        </select>
        Avdeling:
        <select name="avdeling" onchange="form.submit();">
EOF

$q = $dbh->prepare('SELECT DISTINCT(avdeling) FROM fotballserier WHERE sesong=? AND divisjon=? ORDER BY avdeling');
$q->execute($season, $division);

my $found_subdivision = 0;

while (my $ref = $q->fetchrow_hashref) {
	my $sd = $ref->{'avdeling'};
	if ($sd == $subdivision) {
		print "        <option value=\"$sd\" selected=\"selected\">$sd</option>\n";
		$found_subdivision = 1;
	} else {
		print "        <option value=\"$sd\">$sd</option>\n";
	}
}

$subdivision = 1 if (!$found_subdivision);

print <<"EOF";
        </select>
        <input type="submit" value="Vis" />
      </p>
    </form>
EOF

# Get players and ratings
$q = $dbh->prepare('SELECT fotballdeltagere.id,fotballdeltagere.navn,rating FROM fotballdeltagere JOIN fotballserier ON fotballdeltagere.serie=fotballserier.nr LEFT JOIN ratings ON fotballdeltagere.id=ratings.id WHERE sesong=? AND divisjon=? AND avdeling=?');
$q->execute($season, $division, $subdivision);

while (my $ref = $q->fetchrow_hashref) {
	my $id = $ref->{'id'};
	$players{$id} = sanitize(Encode::decode_utf8($ref->{'navn'}));
	$ratings{$id} = $ref->{'rating'};
}
$q->finish;

$q = $dbh->prepare('
SELECT
  d1.id AS p1, d2.id AS p2, maalfor AS score1, maalmot AS score2
FROM fotballresultater r
  JOIN fotballserier s ON r.serie=s.nr
  JOIN fotballdeltagere d1 ON r.lagrecno=d1.nr AND r.serie=d1.serie
  JOIN fotballdeltagere d2 ON r.motstander=d2.nr AND r.serie=d2.serie
WHERE
  sesong=? AND divisjon=? AND avdeling=?
  AND lagrecno > motstander
');
$q->execute($season, $division, $subdivision);

while (my $ref = $q->fetchrow_hashref) {
	push @matches, [ $ref->{'p1'}, $ref->{'p2'}, $ref->{'score1'}, $ref->{'score2'} ];
}
$q->finish;

my $lowest_division = ($division == $max_division);
make_table($lowest_division, {});

print <<"EOF";
    <p>Under er en variant som tar relativ spillestyrke med i beregningen;
      se <a href="/rating">ratingsiden</a>.</p>
EOF

make_table($lowest_division, \%ratings);

print << "EOF";
    </table>
  </body>
</html>
EOF
